
#include "GQPEList.h"
#include "strutils.h"

#include <stdio.h>
#include <strings.h>
#include <stdint.h>
#include <limits.h>
#include <libxml/xmlreader.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>

enum {
  kGQPEListImplicitSingle = 1 << 0
};

typedef struct {
  GQPERef     peListHead;
  int         peFlags;
  char*       peListScriptBase;
} GQPEList;

#define CAST_TO_GQPELIST(P) ((GQPEList*)sllNode_GetMutableDataPtr(P))

int
__GQPEComparator(
  sllNodeRef    pe1,
  sllNodeRef    pe2,
  const void*   context
)
{
  //  Check the identifier:
  if (strcmp(GQPEGetIdentifier(pe1),GQPEGetIdentifier(pe2)) == 0)
    return 0;
  return (GQPEGetCPUCount(pe1) - GQPEGetCPUCount(pe2));
}

//

const xmlChar*    GQPEXMLNo                       = BAD_CAST "no";
const xmlChar*    GQPEXMLFalse                    = BAD_CAST "false";

const xmlChar*    GQPEXMLPEListTag                = BAD_CAST "pe_list";
const xmlChar*    GQPEXMLPEListIDAttr             = BAD_CAST "id";
const xmlChar*    GQPEXMLPEListScriptBaseAttr     = BAD_CAST "script-base";
const xmlChar*    GQPEXMLPEListImplicitSingleAttr = BAD_CAST "implicit-single";

const xmlChar*    GQPEXMLPETag                    = BAD_CAST "pe";
const xmlChar*    GQPEXMLPEIDAttr                 = BAD_CAST "id";
const xmlChar*    GQPEXMLPECPUCountAttr           = BAD_CAST "cpu-count";
const xmlChar*    GQPEXMLPERoundUpAttr            = BAD_CAST "round-up";

//

void
__GQPEListSetUseImplicitSingle(
  GQPEListRef     aPEList,
  int             state
)
{
  GQPEList*   data = CAST_TO_GQPELIST(aPEList);
  
  if (state)
    data->peFlags |= kGQPEListImplicitSingle;
  else
    data->peFlags &= ~kGQPEListImplicitSingle;
}

//

void
__GQPEListXMLErrFunc(
  void*                   arg,
  const char*             msg,
  xmlParserSeverities     severity,
  xmlTextReaderLocatorPtr locator
)
{
  const char *err[5] = {
    "",
    "Validity Warning",
    "Validity Error",
    "Warning",
    "Error"
  };
  fprintf(stderr,"Parse %s [line %i]: %s",err[severity],xmlTextReaderLocatorLineNumber(locator),msg);
}

void
__GQPEListXMLPEListAttr(
  GQPEListRef         gqpes,
  xmlTextReaderPtr    xmlReader
)
{
  xmlChar*      value;
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPEListIDAttr)) ) {
    //  Get the identifier for the PE list:
    GQPEListSetIdentifier(gqpes,(const char*)value);
  }
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPEListScriptBaseAttr)) ) {
    //  Get the script directory for this set of p.e.'s.  We do a canonicalization
    //  here, too:
    char*     canonical = pathstr_canonicalPath((const char*)value);
    
    if (canonical) {
      GQPEListSetScriptBase(gqpes,canonical);
      free(canonical);
    } else {
      printf("ERROR:  Invalid script-base directory: %s\n",(const char*)value);
      exit(1);
    }
  }
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPEListImplicitSingleAttr)) ) {
    //  If value is NOT "no" or "false" then set the flag:
    int   flag = (xmlStrcasecmp(value,GQPEXMLNo) && xmlStrcasecmp(value,GQPEXMLFalse));
    
    __GQPEListSetUseImplicitSingle(gqpes,1);
  }
}

void
__GQPEXMLPEAttr(
  GQPERef             gqpe,
  xmlTextReaderPtr    xmlReader
)
{
  xmlChar*      value;
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPEIDAttr)) ) {
    //  Get the identifier for the PE:
    GQPESetIdentifier(gqpe,(const char*)value);
  }
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPECPUCountAttr)) ) {
    //  If value is valid, then set the CPU count:
    int         count;
    
    if (sscanf((const char*)value,"%d",&count) == 1)
      GQPESetCPUCount(gqpe,count);
  }
  
  if ( (value = xmlTextReaderGetAttribute(xmlReader,GQPEXMLPERoundUpAttr)) ) {
    //  If value is NOT "no" or "false" then set the flag:
    GQPESetRoundUp(gqpe,(xmlStrcasecmp(value,GQPEXMLNo) && xmlStrcasecmp(value,GQPEXMLFalse)));
  }
}

//
#pragma mark -
//

GQPEListRef
GQPEListCreate()
{
  GQPEListRef   newList = sllNode_CreateWithCapacity(sizeof(GQPEList));
  
  if (newList) {
    GQPEList*   data = CAST_TO_GQPELIST(newList);
    
    data->peListHead        = NULL;
    data->peListScriptBase  = NULL;
  }
  return newList;
}

//

GQPEListRef
GQPEListCreateWithXML(
  const char*   filename
)
{
  xmlTextReaderPtr    xmlReader = xmlNewTextReaderFilename(filename);
  GQPEListRef         peListHead = NULL;
  
  if (xmlReader) {
    GQPEListRef       curPEList = NULL;
    GQPERef           curPE = NULL;
    int               done = 0;
    
    xmlTextReaderSetErrorHandler(xmlReader,__GQPEListXMLErrFunc,NULL);

    while ((xmlTextReaderRead(xmlReader) == 1) && (!done)) {
      xmlChar*  name = xmlTextReaderName(xmlReader);
      xmlChar*  value = xmlTextReaderValue(xmlReader);
        
      if (name) {
      
        switch ( xmlTextReaderNodeType(xmlReader) ) {
        
          //  Element begin:
          case XML_READER_TYPE_ELEMENT: {
            //  <pe-list>
            if (xmlStrcasecmp(name,GQPEXMLPEListTag) == 0) {
              //  About to start a new GQPEList, if one was started
              //  we need to add the list to the list list.  Any started
              //  GQPE must be added first:
              if (curPE) {
                if (curPEList)
                  GQPEListAddPE(curPEList,curPE);
                GQPERelease(curPE);
                curPE = NULL;
              }
              if (curPEList) {
                peListHead = GQPEListAppendList(peListHead,curPEList);
                GQPEListRelease(curPEList);
                curPEList = NULL;
              }
              //  Now start the new GQPEList and get attributes:
              if (curPEList = GQPEListCreate())
                __GQPEListXMLPEListAttr(curPEList,xmlReader);
            }
            
            //  <pe>
            else if (xmlStrcasecmp(name,GQPEXMLPETag) == 0 && curPEList) {
              //  About to start a new GQPE, if one was started
              //  and there's an existing GQPEList, add the older GQPE
              //  to that GQPEList before proceeding:
              if (curPE) {
                if (curPEList)
                  GQPEListAddPE(curPEList,curPE);
                GQPERelease(curPE);
                curPE = NULL;
              }
              //  Now start the new GQPE and get attributes:
              if (curPE = GQPECreateWithIDAndCPUCount(GQPEListGetIdentifier(curPEList),0)) {
                __GQPEXMLPEAttr(curPE,xmlReader);
                //  If implicit single is on for this list, then no single-processor
                //  PE is allowed!
                if (GQPEListUseImplicitSingle(curPEList) && (GQPEGetCPUCount(curPE) == 1)) {
                  GQPERelease(curPE);
                  curPE = NULL;
                }
              }
            }
            break;
          }
          
          case XML_READER_TYPE_END_ELEMENT: {
            //  </pe-list>
            if (xmlStrcasecmp(name,GQPEXMLPEListTag) == 0) {
              //  Ending a new GQPEList, if one was started
              //  we need to add the list to the list list.  Any started
              //  GQPE must be added first:
              if (curPE) {
                if (curPEList)
                  GQPEListAddPE(curPEList,curPE);
                GQPERelease(curPE);
                curPE = NULL;
              }
              if (curPEList) {
                peListHead = GQPEListAppendList(peListHead,curPEList);
                GQPEListRelease(curPEList);
                curPEList = NULL;
              }
            }
            
            //  </pe>
            else if (xmlStrcasecmp(name,GQPEXMLPETag) == 0) {
              //  Ending a new GQPE, if one was started
              //  and there's an existing GQPEList, add the older GQPE
              //  to that GQPEList before proceeding:
              if (curPE) {
                if (curPEList) {
                  //  If implicit single is on for this list, then no single-processor
                  //  PE is allowed!
                  if (GQPEListUseImplicitSingle(curPEList) && (GQPEGetCPUCount(curPE) == 1)) {
                    GQPERelease(curPE);
                    curPE = NULL;
                  } else
                    GQPEListAddPE(curPEList,curPE);
                }
                GQPERelease(curPE);
                curPE = NULL;
              }
            }
            break;
          }
        
        }
      }
    }
    
    //  If we get here and there's a remaining GQPE and GQPEList, add
    //  them:
    if (curPE) {
      if (curPEList) {
        //  If implicit single is on for this list, then no single-processor
        //  PE is allowed!
        if (GQPEListUseImplicitSingle(curPEList) && (GQPEGetCPUCount(curPE) == 1)) {
          GQPERelease(curPE);
          curPE = NULL;
        } else
          GQPEListAddPE(curPEList,curPE);
      }
      GQPERelease(curPE);
      curPE = NULL;
    }
    if (curPEList) {
      peListHead = GQPEListAppendList(peListHead,curPEList);
      GQPEListRelease(curPEList);
      curPEList = NULL;
    }
    xmlFreeTextReader(xmlReader);
  }
  return peListHead;
}

//

GQPEListRef
GQPEListRetain(
  GQPEListRef     aPEList
)
{
  return sllNode_Retain(aPEList);
}

//

void
GQPEListRelease(
  GQPEListRef     aPEList
)
{
  if (sllNode_GetReferenceCount(aPEList) == 1) {
    //  Time to deallocate:
    GQPEList*     data = CAST_TO_GQPELIST(aPEList);
    
    if (data->peListHead)
      sllNode_Release(data->peListHead);
    if (data->peListScriptBase)
      free(data->peListScriptBase);
  }
  sllNode_Release(aPEList);
}

//

const char*
GQPEListGetIdentifier(
  GQPEListRef     aPEList
)
{
  return sllNode_GetIdentifier(aPEList);
}

//

void
GQPEListSetIdentifier(
  GQPEListRef     aPEList,
  const char*     peListID
)
{
  sllNode_SetIdentifier(aPEList,peListID);
  strtolower((char*)sllNode_GetIdentifier(aPEList));
}

//

const char*
GQPEListGetScriptBase(
  GQPEListRef     aPEList
)
{
  return CAST_TO_GQPELIST(aPEList)->peListScriptBase;
}

//

void
GQPEListSetScriptBase(
  GQPEListRef     aPEList,
  const char*     scriptBase
)
{
  GQPEList*       data = CAST_TO_GQPELIST(aPEList);
  
  if (data->peListScriptBase)
    free(data->peListScriptBase);
  if (scriptBase && strlen(scriptBase))
    data->peListScriptBase = strdup(scriptBase);
  else
    data->peListScriptBase = NULL;
}

//

char*
GQPEListMakeScriptPath(
  GQPEListRef aPEList,
  int         which
)
{
  GQPEList*   data = CAST_TO_GQPELIST(aPEList);
  char*       path = NULL;
  
  switch (which) {
    case kGQPEListPreMergeScript:
      path = strcatm(data->peListScriptBase,"/pre",NULL);
      break;
      
    case kGQPEListValidatePEScript:
      path = strcatm(data->peListScriptBase,"/validate-pe",NULL);
      break;
      
    case kGQPEListMergeInputScript:
      path = strcatm(data->peListScriptBase,"/merge",NULL);
      break;
    
    case kGQPEListPostMergeScript:
      path = strcatm(data->peListScriptBase,"/post",NULL);
      break;
    
  }
  
  //  If we made a script path then test whether it exists and is executable:
  if (path) {
    if (!pathstr_fileExists(path)) {
      free(path);
      path = NULL;
    } else if (pathstr_isDirectory(path)) {
      free(path);
      path = NULL;
    } else if (!pathstr_isExecutable(path)) {
      free(path);
      path = NULL;
    }
  }
  
  return path;
}

//

int
GQPEListUseImplicitSingle(
  GQPEListRef     aPEList
)
{
  GQPEList*       data = CAST_TO_GQPELIST(aPEList);
  return ((data->peFlags & kGQPEListImplicitSingle) != 0);
}

//

void
GQPEListAddPE(
  GQPEListRef     aPEList,
  GQPERef         aPE
)
{
  GQPEList*       data = CAST_TO_GQPELIST(aPEList);
  
  if (data->peListHead)
    data->peListHead = sllNode_SortedInsertion(data->peListHead,aPE,__GQPEComparator,NULL,0);
  else
    data->peListHead = GQPERetain(aPE);
}

//

void
GQPEListRemovePE(
  GQPEListRef     aPEList,
  GQPERef         aPE
)
{
  GQPEList*       data = CAST_TO_GQPELIST(aPEList);
  
  if (data->peListHead)
    data->peListHead = sllNode_RemoveNode(data->peListHead,aPE);
}

//

unsigned int
__GQPEValidForNProc(
  GQPERef         gqpe,
  unsigned int    nproc
)
{
  if (gqpe) {
    unsigned int  cpuCount = GQPEGetCPUCount(gqpe);
    
    if (nproc <= cpuCount) {
      if (GQPEGetRoundUp(gqpe))
        return cpuCount;
      return nproc;
    }
  }
  return 0;
}

GQPERef
GQPEListSelectPE(
  GQPEListRef     aPEList,
  unsigned int*   nproc
)
{
  GQPERef         gqpe = CAST_TO_GQPELIST(aPEList)->peListHead;
  
  while (gqpe) {
    int     adjNProc = __GQPEValidForNProc(gqpe,*nproc);
    
    if (adjNProc) {
      *nproc = adjNProc;
      break;
    }
    gqpe = sllNode_GetLink(gqpe);
  }
  return gqpe;
}

//

unsigned int
GQPEListMaximumCPUCount(
  GQPEListRef     aPEList
)
{
  GQPERef         gqpe = CAST_TO_GQPELIST(aPEList)->peListHead;
  unsigned int    count = 0;
  
  while (gqpe) {
    unsigned int  tmp = GQPEGetCPUCount(gqpe);
    
    if (tmp > count)
      count = tmp;
    gqpe = sllNode_GetLink(gqpe);
  }
  return count;
}

//

void
GQPEListSummarize(
  GQPEListRef     aPEList
)
{
  GQPEList*       data = NULL;
  
  if ( aPEList && (data = CAST_TO_GQPELIST(aPEList))) {
    GQPERef       gqpe = data->peListHead;
    const char*   listName = GQPEListGetIdentifier(aPEList);
    
    printf(
      "------------------------------------------------------------------\n"
      "%-20s%46s\n"
      "+---------------+-----+------------------------------------------+\n"
      " ID               CPU   Flags\n"
      "+---------------+-----+------------------------------------------+\n",
      (listName ? listName : "unnamed"),
      (data->peListScriptBase ? data->peListScriptBase : "no script base")
    );
    if (gqpe) {
      while (gqpe) {
        GQPESummarize(gqpe);
        gqpe = sllNode_GetLink(gqpe);
      }
    } else
      printf(" No parallel environments available.\n");
    printf("------------------------------------------------------------------\n");
  }
}

//

void
GQPEListSummarizeAll(
  GQPEListRef     aPEList
)
{
  while (aPEList) {
    GQPEListSummarize(aPEList);
    if (aPEList = sllNode_GetLink(aPEList))
      printf("\n");
  }
}

//
#pragma mark -
//

GQPEListRef
GQPEListAppendList(
  GQPEListRef     aPEList,
  GQPEListRef     altPEList
)
{
  return sllNode_AppendNode(aPEList,altPEList);
}

//

GQPEListRef
GQPEListRemoveList(
  GQPEListRef     aPEList,
  GQPEListRef     targetPEList
)
{
  return sllNode_RemoveNode(aPEList,targetPEList);
}

//

GQPEListRef
GQPEListGetListWithIdentifier(
  GQPEListRef     aPEList,
  const char*     peListID
)
{
  return sllNode_Search(aPEList,sllNodeIdentifierSearcher,peListID);
}
