//
//  gqueue
//
//  Program which creates SGE queue scripts for various types of
//  computational programs.  This is the second version, which moves
//  the parallel environment specification into an external XML file
//  rather than compiling it into the code.  It also makes the process
//  somewhat more free-form, offloading some of the "special" script-
//  generating stuff to scripts/executables that the system admin
//  provides for each p.e. type:  grep'ing a processor count from the
//  Gaussian input, for example, is accomplished externally in a
//  pre-input-merge script.  The fancy shell scripting that selects
//  the SMP or TCPLinda version of Gaussian is embedded in the
//  script that merges the input file (and must remove any nproc
//  lines) into the qsub script.
//
//  By default the configuration and scripts reside in /etc/gqueue.d.
//
//  Created by Jeffrey Frey on Tue Jan 10 2006.
//  Network & Systems Services
//  University of Delaware
//  Copyright (c) 2006
//
//  Modifications:
//
//  13 Mar 2006 [JTF]
//  Added priority specification capability.
//
//  13 Jul 2006 [JTF]
//  Added a hook for another script to modify the parallel enviroment
//  name if necessary.  I put this in mainly because of the group-based
//  queues on spongebob.  Also changed the script names to eliminate the
//  need for the PE entity name -- e.g. now just "pre" instead of
//  "gaussian.pre".  Not really necessary, since we promote the use of
//  directories to group scripts on a per-PE-type basis.
//

#include "gqueue.h"

enum {
  /* Version Release Stage Codes */
  developStage                  = 0x20,
  alphaStage                    = 0x40,
  betaStage                     = 0x60,
  finalStage                    = 0x80
};

//
// Version => 2.0.1 f 0 = 0x02018000
//
#ifndef PROGRAM_VERSION
#define PROGRAM_VERSION ( 2 << 24 | 1 << 20 | 0 << 16 | finalStage << 8 | 0 )
#endif

const int     gqTrue = 1;
const int     gqFalse = 0;

//
// The default path to gqueue configuration data:
//
const char*   gqConfigPath = "/etc/gqueue.d";
const char*   gqPEConfigPath = "/etc/gqueue.d/pe.conf";
const char*   gqDefaultQScriptDir = "~/.gqueue";

//
// The priority descriptors:
//
#define GQPRIORITY_LOW "low"
#define GQPRIORITY_MEDIUM "medium"
#define GQPRIORITY_HIGH "high"
const char*   gqPriority_Low = GQPRIORITY_LOW;
const char*   gqPriority_Medium = GQPRIORITY_MEDIUM;
const char*   gqPriority_High = GQPRIORITY_HIGH;
enum {
  kGQPriority_LowValue    =    0,
  kGQPriority_MediumValue =  +50,
  kGQPriority_HighValue   = +100
};

//
// These globals form the gqueue runtime state:
//
int           gqVerbose = 0;                  static const char*  GQVerboseEnv = "GQUEUE_VERBOSE";
int           gqDebug = 0;
int           gqHelp = 0;
int           gqSaveOutput = 0;               static const char*  GQSaveOutputEnv = "GQUEUE_SAVEOUTPUT";
int           gqSaveScript = 0;               static const char*  GQSaveScriptEnv = "GQUEUE_SAVESCRIPT";
int           gqPESummary = 0;
char*         gqEmailAddress = NULL;          static const char*  GQEmailAddressEnv = "GQUEUE_EMAILADDR";
char*         gqQueueDirectory = NULL;        static const char*  GQQueueDirEnv = "GQUEUE_QUEUEDIR";
char*         gqShellPath = NULL;             static const char*  GQShellPathEnv = "GQUEUE_SHELLPATH";
unsigned int  gqNProc = 0;                    static const char*  GQNProcEnv = "GQUEUE_NPROC";
char*         gqInputFilePath = NULL;         static const char*  GQInputFileEnv = "GQUEUE_INPUT_FILEPATH";
char*         gqScriptFilePath = NULL;        static const char*  GQScriptFileEnv = "GQUEUE_SCRIPT_FILEPATH";
const char*   gqPriority = GQPRIORITY_MEDIUM; static const char*  GQPriorityEnv = "GQUEUE_PRIORITY";
unsigned int  gqMem = 0;                      static const char*  GQMemSizeEnv = "GQUEUE_MEMSIZE";

static const char*  GQPENameEnv = "GQUEUE_PE_NAME";
static const char*  GQPETypeEnv = "GQUEUE_PE_TYPE";

int           gqDeleteInputFile = 0;

//
// This is our generic exit point, which gets rid of any files
// we don't want to hang around, etc:
//
void
GQExit(
  int   status
)
{
  if (gqDeleteInputFile)
    unlink(gqInputFilePath);
  
  //  Remove the script:
  if (gqScriptFilePath && !gqSaveScript)
    unlink(gqScriptFilePath);
  
  exit(status);
}

//
// The following section gives us a string-to-enum
// mapping of environment variable names:
//
enum {
  kGQVerboseEnvVar        = 0,
  kGQSaveOutputEnvVar     = 1,
  kGQSaveScriptEnvVar     = 2,
  kGQEmailAddressEnvVar   = 3,
  kGQNProcEnvVar          = 4,
  kGQMemSizeEnvVar        = 5,
  kGQShellPathEnvVar      = 6,
  kGQPENameEnvVar         = 7,
  kGQMaxEnvVar            = 8
};
const char** gqImportableEnvVars[kGQMaxEnvVar] = {
  &GQVerboseEnv,
  &GQSaveOutputEnv,
  &GQSaveScriptEnv,
  &GQEmailAddressEnv,
  &GQNProcEnv,
  &GQMemSizeEnv,
  &GQShellPathEnv,
  &GQPENameEnv
};
int
GQTranslateEnvVarName(
  char*     vName
)
{
  int       i = 0;
  while ( i < kGQMaxEnvVar ) {
    if (strcmp(*(gqImportableEnvVars[i]),vName) == 0)
      return i;
    i++;
  }
  return -1;
}

//

size_t
GQParseMemorySpec(
  char*   memstr
)
{
  size_t  mem = 0;
  int     nchar;
  
  if (*memstr && (sscanf(memstr,"%u%n",&mem,&nchar) == 1) && (mem > 0)) {
    char*       label = memstr + nchar;
    double      multiplier = 1.0;
    
    while ( *label && isspace(*label) ) label++;
    if ( *label ) {
      switch ( *label ) {
        case 'k':
          multiplier = 1000.0; break;
        case 'K':
          multiplier = 1024.0; break;
        case 'm':
          multiplier = 1000000.0; break;
        case 'M':
          multiplier = 1048576.0; break;
        case 'g':
          multiplier = 1000000000.0; break;
        case 'G':
          multiplier = 1073741824.0; break;
        default:
          fprintf(stderr,"ERROR:  Invalid memory size specification: %s\n",memstr);
          GQExit(1);
      }
      if ( *(++label) ) {
        switch ( *label ) {
          case 'b':
          case 'B':
            break;
          case 'w':
          case 'W':
            multiplier *= 8.0;
          default:
            fprintf(stderr,"ERROR:  Invalid memory size specification: %s\n",memstr);
            GQExit(1);
        }
      }
    }
    mem = ceil( (mem * multiplier) / 1048576.0 );
  }
  return mem;
}

//
// We wouldn't necessarily have to make the selected PE list and PE
// global, but we'll never need more than one of each within the
// context of this program:
//
static GQPEListRef  gqTargetPEList = NULL;
static GQPERef      gqTargetPE = NULL;

//
// The extended getopt() parameter block:
//
static struct option gqOptions[] =
{
  { "verbose",      no_argument,        NULL,           'v' },
  { "version",      no_argument,        NULL,           'V' },
  { "quiet",        no_argument,        NULL,           'q' },
  { "debug",        no_argument,        NULL,           'd' },
  { "help",         no_argument,        &gqHelp,         1  },
  { "saveoutput",   no_argument,        &gqSaveOutput,   1  },
  { "savescript",   no_argument,        &gqSaveScript,   1  },
  { "pe-summary",   no_argument,        &gqPESummary,    1  },
  { "queuedir",     required_argument,  NULL,           'Q' },
  { "email-address",required_argument,  NULL,           'e' },
  { "nproc",        required_argument,  NULL,           'n' },
  { "pe-type",      required_argument,  NULL,           't' },
  { "priority",     required_argument,  NULL,           'p' },
  { "memsize",      required_argument,  NULL,           'm' },
  { "shell",        required_argument,  NULL,           's' },
  { NULL,           0,                  NULL,            0  }
};

//
#pragma mark -
//

/*!
  @function printvers
  Turns a 32-bit version "string" into a textual representation and
  spits it out on the given <TT>stream</TT>.
*/
void
printvers(
  FILE*       stream,
  u_int32_t   vers
)
{
  short               major,minor,bug,stage,nonRel;
  
  major = (vers & 0xFF000000) >> 24;
  minor = (vers & 0x00F00000) >> 20;
  bug = (vers & 0x000F0000) >> 16;
  stage = (vers & 0x0000FF00) >> 8;
  nonRel = (vers & 0x000000FF);
  
  fprintf(stream,"%d.%d",major,minor);
  if (bug)
    fprintf(stream,".%d",bug);
  switch (stage) {
    case developStage:
      fputc('d',stream);
      break;
    case alphaStage:
      fputc('a',stream);
      break;
    case betaStage:
      fputc('b',stream);
      break;
    case finalStage:
      fputc('f',stream);
      break;
  }
  if (nonRel)
    fprintf(stream,"%d",nonRel);
}

/*!
  @function usage
  Print the usage/help page for the program.  The executable (from
  argv[0]) must be passed to this routine.
*/
void
usage(
  const char*     executable
)
{
  printf("usage [");
  printvers(stdout,PROGRAM_VERSION);
  printf("]:\n\n"
    "  %s {options} <input-file>\n\n"
    "  options:\n"
    "    --verbose,-v         Write additional information as we run.\n"
    "    --quiet,-q           Write as little information as possible\n"
    "                         as we run.\n"
    "    --help               Display this information.\n"
    "    --version,-V         Just show the version and compile date.\n"
    "    --saveoutput         Retain SGE output file after the job completes.\n"
    "    --savescript         Retain the SGE submission script after the\n"
    "                         job completes.\n"
    "    --queuedir,-Q        Provide an alternate location for SGE files\n"
    "                         (must be accompanied by a file path).\n"
    "    --email-address,-e   Send an email to this address when the\n"
    "                         job dies/finishes.\n"
    "    --nproc,-n           Number of processors to request for the job.\n"
    "    --memsize,-m         Amount of memory to request for the job.  Valid\n"
    "                         labels are:  {k,K,m,M,g,G}{b,B,w,W}\n"
    "    --pe-type,-t         Select the indicated type of parallel environment\n"
    "                         for this job.\n"
    "    --pe-summary         Summarize the selected p.e. type or if no p.e.\n"
    "                         type is selected, all p.e. types.\n"
    "    --priority,-p        Indicate the initial scheduling priority for the\n"
    "                         job:  low, medium, or high (default: medium)\n"
    "    --shell,-s           Path to the shell we want this job to run in\n"
    "  example:\n"
    "    %s --email-address=user@host.com --nproc=4 --pe-type=gaussian\n"
    "    %s --email-address=user@host.com -n 8 -p=low --verbose\n\n",
    executable,executable,executable
  );
}

//

/*!
  @function GQStringToBoolean
  Attempt to gleen a boolean value from an environment variable.  Values that
  yield boolean true are:
  <UL>
    <LI>yes</LI>
    <LI>true</LI>
    <LI>non-zero integer</LI>
    <LI>empty string</LI>
  </UL>
  Note than an empty string is NOT the same as the value's coming back a
  <TT>NULL</TT> pointer.
*/
int
GQStringToBoolean(
  const char* envVar
)
{
  char*     str;
  int       value;
  if (str = getenv(envVar)) {
    if (*str == '\0')
      return gqTrue;
    if (strcasecmp(str,"yes") == 0)
      return gqTrue;
    if (strcasecmp(str,"true") == 0)
      return gqTrue;
    if ((sscanf(str,"%d",&value) == 1) && (value))
      return gqTrue;
  }
  return gqFalse;
}

/*!
  @function GQBooleanToString
  If <TT>bval</TT> is zero then the string "no" is returned.  Otherwise,
  a "yes" string is returned.  The string is constant, so the caller
  should not attempt to modify or free it.
*/
const char*
GQBooleanToString(
  int     bval
)
{
  return ( bval ? "yes" : "no" );
}

/*!
  @function GQExportEnv
  Exports all of the runtime state variables to the environment, so that
  scripts can get state information.
*/
void
GQExportEnv()
{
  char              nproc[16];
  
  if (gqVerbose)
    setenv(GQVerboseEnv, "1" , 1);
  else
    unsetenv(GQVerboseEnv);
  
  if (gqSaveOutput)
    setenv(GQSaveOutputEnv, "1" , 1);
  else
    unsetenv(GQSaveOutputEnv);
  
  if (gqSaveScript)
    setenv(GQSaveScriptEnv, "1" , 1);
  else
    unsetenv(GQSaveScriptEnv);
  
  if (gqEmailAddress)
    setenv(GQEmailAddressEnv, gqEmailAddress , 1);
  else
    unsetenv(GQEmailAddressEnv);
    
  if (gqShellPath)
    setenv(GQShellPathEnv, gqShellPath , 1);
  else
    unsetenv(GQShellPathEnv);
  
  if (gqQueueDirectory)
    setenv(GQQueueDirEnv, gqQueueDirectory , 1);
  else
    unsetenv(GQQueueDirEnv);
  
  if (gqNProc > 0) {
    snprintf(nproc,16,"%u",gqNProc);
    setenv(GQNProcEnv,nproc,1);
  }
  
  if (gqMem > 0) {
    snprintf(nproc,16,"%uMB",gqMem);
    setenv(GQMemSizeEnv,nproc,1);
  }
  
  if (gqPriority)
    setenv(GQPriorityEnv,gqPriority,1);
  else
    unsetenv(GQPriorityEnv);
  
  if (gqInputFilePath)
    setenv(GQInputFileEnv, gqInputFilePath , 0); // overwrite = 0, so that we only set it ONCE
  else
    unsetenv(GQInputFileEnv);
  
  if (gqScriptFilePath)
    setenv(GQScriptFileEnv, gqScriptFilePath , 0); // overwrite = 0, so that we only set it ONCE
  else
    unsetenv(GQScriptFileEnv);
    
  if ( gqTargetPE )
    setenv(GQPENameEnv, GQPEGetIdentifier(gqTargetPE), 0);
  else
    unsetenv(GQPENameEnv);
}

/*!
  @function GQImportEnv
  Imports runtime state information from the environment.
*/
void
GQImportEnv()
{
  char*         value;
  
  gqVerbose = GQStringToBoolean(GQVerboseEnv);
  gqSaveOutput = GQStringToBoolean(GQSaveOutputEnv);
  gqSaveScript = GQStringToBoolean(GQSaveScriptEnv);
  
  if ( value = getenv(GQEmailAddressEnv) ) {
    if (gqEmailAddress) {
      if (strcmp(gqEmailAddress,value)) {
        free(gqEmailAddress);
        gqEmailAddress = strdup(value);
      }
    } else
      gqEmailAddress = strdup(value);
  } else if (gqEmailAddress) {
    free(gqEmailAddress);
    gqEmailAddress = NULL;
  }
  
  if ( value = getenv(GQQueueDirEnv) ) {
    if (gqQueueDirectory) {
      if (strcmp(gqQueueDirectory,value)) {
        free(gqQueueDirectory);
        gqQueueDirectory = strdup(value);
      }
    } else
      gqQueueDirectory = strdup(value);
  } else if (gqQueueDirectory) {
    free(gqQueueDirectory);
    gqQueueDirectory = NULL;
  }
  
  if ( value = getenv(GQShellPathEnv) ) {
    if ( gqShellPath ) free(gqShellPath);
    gqShellPath = strdup(value);
  }
  
  if ( value = getenv(GQNProcEnv) ) {
    unsigned int      nproc;
    
    if ((sscanf(value,"%u",&nproc) == 1) && (nproc > 0))
      gqNProc = nproc;
  }
  
  if ( value = getenv(GQMemSizeEnv) ) {
    unsigned int      mem = GQParseMemorySpec(value);
    
    if ( mem && mem > gqMem )
      gqMem = mem;
  }
  
  if ( value = getenv(GQPriorityEnv) ) {
    if ( strcasecmp(value,gqPriority_Low) == 0 ) {
      gqPriority = gqPriority_Low;
    } else if ( strcasecmp(value,gqPriority_Medium) == 0 ) {
      gqPriority = gqPriority_Medium;
    } else if ( strcasecmp(value,gqPriority_High) == 0 ) {
      gqPriority = gqPriority_High;
    }
  }
  
  if ( !gqTargetPEList && (value = getenv(GQPETypeEnv)) ) {
    //  Attempt to select the p.e. type:
    strtolower(value);
    if ((gqTargetPEList = GQPEListGetListWithIdentifier(gqTargetPEList,optarg)) == NULL) {
      printf("ERROR:  The PE type '%s' does not exist.\n",optarg);
      GQExit(1);
    }
  }
  
  if (gqDebug) {
    //  Summarize the import of the environment variables:
    fprintf(stderr,
        ":: gqueue: imported state from environment:\n"
        "::   %20s = %s\n"
        "::   %20s = %s\n"
        "::   %20s = %s\n"
        "::   %20s = %s\n"
        "::   %20s = %s\n"
        "::   %20s = %u\n"
        "::   %20s = %uMB\n"
        "::   %20s = %s\n"
        "::   %20s = %s\n",
        GQVerboseEnv,GQBooleanToString(gqVerbose),
        GQSaveOutputEnv,GQBooleanToString(gqSaveOutput),
        GQSaveScriptEnv,GQBooleanToString(gqSaveScript),
        GQEmailAddressEnv,(gqEmailAddress ? gqEmailAddress : "<not set>"),
        GQQueueDirEnv,(gqQueueDirectory ? gqQueueDirectory : "<not set>"),
        GQNProcEnv,gqNProc,
        GQMemSizeEnv,gqMem,
        GQPriorityEnv,gqPriority,
        GQShellPathEnv,gqShellPath
    );
  }
}

/*!
  @function GQIntegerPriorityForString
  Return an integer priority value for the string passed to us.
*/
int
GQIntegerPriorityForString(
  const char*   priorityStr
)
{
  int     priority = kGQPriority_MediumValue;
  
  if ( priorityStr ) {
    if ( strcasecmp(priorityStr,gqPriority_Low) == 0 ) {
      priority = kGQPriority_LowValue;
    } else if ( strcasecmp(priorityStr,gqPriority_Medium) == 0 ) {
      priority = kGQPriority_MediumValue;
    } else if ( strcasecmp(priorityStr,gqPriority_High) == 0 ) {
      priority = kGQPriority_HighValue;
    }
  }
  return priority;
}

//
#pragma mark -
//

/*!
  @function GQReadInputFileFromStdin
  If the user did not provide an input filename on the command line, then
  we end up reading it from <TT>stdin</TT>.  This routine dumps the entirety
  of <TT>stdin</TT> to a temporary file that becomes our input file.
*/
char*
GQReadInputFileFromStdin()
{
  //  Open a temporary file and dump stdin to that:
  char*     tmpName = strdup("/tmp/gqueue.inXXXXXX");
  char      buffer[512];
  FILE*     fptr;
  
  mktemp(tmpName);
  if (fptr = fopen(tmpName,"w")) {
    while (!feof(stdin)) {
      size_t    bytes = fread(buffer,1,512,stdin);
      
      if (bytes > 0)
        fwrite(buffer,1,bytes,fptr);
    }
    fclose(fptr);
    gqDeleteInputFile = gqTrue;
    return tmpName;
  }
  return NULL;
}

/*!
  @function GQRunScript
  Runs the script/executable at <TT>scriptPath</TT>.  This is for the external
  scripts/executables that gqueue calls to do any extra SGE qsub script
  setup.  Such scripts output to their <TT>stdout</TT> new values for environment
  variables with lines like "GQUEUE_NPROC=4" and verbose comments come in prefixed
  with a pound sign, a'la "#Found nproc=4 in H2.com".  We grab the script/executable
  <TT>stdout</TT> with a pipe and process that output.
  @param allowVarReset Pass <TT>gqFalse</TT> if incoming variable assignment lines
    should NOT be processed
*/
int
GQRunScript(
  char*   scriptPath,
  int     allowVarReset
)
{
  FILE*   outPipe = popen(scriptPath,"r");
  
  if (outPipe) {
    //  Read the pipe until an EOF is hit:
    while (!feof(outPipe)) {
      size_t  lineLen;
      char*   p = fgetln(outPipe,&lineLen);
      
      if (p) {
        //  Zap any leading whitespace:
        while ( lineLen && isspace(*p) ) {
          p++;lineLen--;
        }
        //  Comment line?
        if (*p == '#' && gqVerbose) {
          if (p[lineLen - 1] == '\n') {
            p[lineLen - 1] = '\0';
            memmove(p,p + 1,lineLen - 1);
          } else {
            memmove(p,p + 1,lineLen - 1);
            p[lineLen - 1] = '\0';
          }
          logtostream(stdout,p);
        } else if (allowVarReset) {
          int     l = 0;
          int     vID;
          char*   pp = p;
          
          //  Figure out the limits for the variable name:
          while ( *pp != '=' && l < lineLen ) {
            pp++;l++;
          }
          if ( l < lineLen ) {
            *pp = '\0';
            lineLen -= (l + 1);
            if ((vID = GQTranslateEnvVarName(p)) >= 0) {
              //  Valid variable, look at it's value:
              char      value[lineLen + 1];
              
              //  Get rid of any trailing whitespace (include newlines):
              ++pp;
              while ( lineLen && isspace(pp[lineLen - 1]) ) lineLen--;
              if (lineLen)
                memcpy(value,pp,lineLen);
              value[lineLen] = '\0';
              
              switch (vID) {
                case kGQVerboseEnvVar:
                  gqVerbose = GQStringToBoolean(value);
                  break;
                case kGQSaveOutputEnvVar:
                  gqSaveOutput = GQStringToBoolean(value);
                  break;
                case kGQSaveScriptEnvVar:
                  gqSaveScript = GQStringToBoolean(value);
                  break;
                case kGQEmailAddressEnvVar: {
                  if (gqEmailAddress) free(gqEmailAddress);
                  if (strlen(value))
                    gqEmailAddress = strdup(gqEmailAddress);
                  else
                    gqEmailAddress = NULL;
                  break;
                }
                case kGQNProcEnvVar: {
                  unsigned int  nproc;
                  
                  if ((sscanf(value,"%u",&nproc) == 1) && (nproc > 0))
                    gqNProc = nproc;
                  break;
                }
                case kGQMemSizeEnvVar: {
                  unsigned int  mem = GQParseMemorySpec(value);

                  if ( mem && mem > gqMem )
                    gqMem = mem;
                  break;
                }
                case kGQShellPathEnvVar: {
                  if ( gqShellPath ) free(gqShellPath);
                  gqShellPath = strdup(value);
                  break;
                }
                case kGQPENameEnvVar: {
                  setenv(GQPENameEnv,value,1);
                  break;
                }
              }
              
            }
          }
        }
      }
    }
    fclose(outPipe);
  }
  return 0;
}

/*!
  @function GQSetupScriptDirectory
  Creates the gqueue script/output directory if it does not exist.
*/
char*
GQSetupScriptDirectory()
{
  char*       path = pathstr_canonicalPath(gqQueueDirectory);
  
  if (path) {
    if (pathstr_fileExists(path)) {
      if (!pathstr_isDirectory(path)) {
        logtostream(stderr,"%s is not a directory",gqQueueDirectory);
        free(path);
        path = NULL;
      }
    } else if (mkdir(path,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)) {
      logtostream(stderr,"%s could not be created",path);
      free(path);
      path = NULL;
    }
  }
  return path;
}

/*!
  @function GQCreateScriptFile
  Creates the SGE qsub script, adding the shell specifier and gqueue
  versioning information.  Returns the path to the qsub script -- the
  caller has ownership of the path and is responsible for releasing
  it.
*/
char*
GQCreateScriptFile()
{
  //  Open a temporary file in the script directory:
  char*     sTmp1;
  char*     sScriptFile = NULL;
  
  if (sTmp1 = pathstr_basenameForPath(gqInputFilePath)) {
    //  SGE does NOT like numbers at the start of a queue script name:
    if (!isdigit(sTmp1[0])) {
      sScriptFile = strcatm(gqQueueDirectory,"/",sTmp1,".gqs_XXXXXXXX",NULL);free(sTmp1);
    } else {
      sScriptFile = strcatm(gqQueueDirectory,"/","gqs_",sTmp1,".XXXXXXXX",NULL);free(sTmp1);
    }
    if (mktemp(sScriptFile)) {
      FILE*         scriptStream;
      
      if (gqVerbose)
        logtostream(stdout,"Job will use queue script: %s",sScriptFile);
      if (scriptStream = fopen(sScriptFile,"w")) {
        //  Set the executable bit:
        chmod(sScriptFile,S_IRWXU);
      
        //  Write-out the comment/header to the script:
        fprintf(scriptStream,"#!%s\n#\n# Script generated by GQueue ",( gqShellPath ? gqShellPath : "/bin/sh" ));
        printvers(scriptStream,PROGRAM_VERSION);
        fprintf(scriptStream,"\n#\n");
        
        //  Close the file and return:
        fclose(scriptStream);
      } else {
        free(sScriptFile);
        sScriptFile = NULL;
      }
    } else {
      free(sScriptFile);
      sScriptFile = NULL;
    }
  }
  return sScriptFile;
}

/*!
  @function GQScriptSGESection
  Writes the SGE control lines to the qsub script.
*/
void
GQScriptSGESection()
{
  FILE*   scriptStream = fopen(gqScriptFilePath,"a+");
    
  if (scriptStream) {
    fprintf(scriptStream,"#$ -S %s\n",( gqShellPath ? gqShellPath : "/bin/sh" ));
    fprintf(scriptStream,"#$ -j y -r n\n");
    fprintf(scriptStream,"#$ -o %s\n",gqQueueDirectory);
    if (gqEmailAddress) {
      fprintf(scriptStream,"#$ -m eas\n");
      fprintf(scriptStream,"#$ -M %s\n",gqEmailAddress);
    }
    fprintf(scriptStream,"#$ -cwd\n");
    if ( ((gqNProc == 1) && !GQPEListUseImplicitSingle(gqTargetPEList)) || gqTargetPE ) {
      char*     nameFromEnv = getenv(GQPENameEnv);
      
      fprintf(scriptStream,"#$ -pe %s %u\n",(nameFromEnv ? nameFromEnv : GQPEGetIdentifier(gqTargetPE)),gqNProc);
    }
    if ( gqMem > 0 )
      fprintf(scriptStream,"#$ -l mf=%uM\n",gqMem);
    fclose(scriptStream);
  }
}

/*!
  @function GQScriptMergeInput
  Merge the input file into the qsub script.  If the selected PE list has a
  script/executable for this purpose then it gets used; otherwise, we do a basic
  file dump ourselves.
*/
void
GQScriptMergeInput()
{
  FILE*       scriptStream = fopen(gqScriptFilePath,"a+");
  
  if (scriptStream) {
    char*     scriptPath = GQPEListMakeScriptPath(gqTargetPEList,kGQPEListMergeInputScript);
    int       result;
    
    //  Is there a script to call?
    if (scriptPath) {
      GQExportEnv();
      if (result = GQRunScript(scriptPath,gqFalse)) {
        printf("ERROR:  %s returned error code %d\n",scriptPath,result);
        GQExit(result);
      }
      free(scriptPath);
    } else {
      //  Built-in merge:
      FILE*   inputFile = fopen(gqInputFilePath,"r");
      char    buffer[512];
      
      if (inputFile) {
        while (!feof(inputFile)) {
          size_t  bytes = fread(buffer,1,512,inputFile);
          
          if (bytes)
            fwrite(buffer,1,bytes,scriptStream);
        }
        fclose(inputFile);
      }
    }
    
    fclose(scriptStream);
  }
}

/*!
  @function GQScriptFileFooter
  Write the closing portion of the qsub script.
*/
void
GQScriptFileFooter()
{
  if ((!gqSaveOutput) || (!gqSaveScript)) {
    FILE*       scriptStream = fopen(gqScriptFilePath,"a+");
    
    if (scriptStream) {
      //  Delete the script when it finishes?
      if (!gqSaveScript)
        fprintf(scriptStream,"rm -f %s\n\n",gqScriptFilePath);
      
      //  Delete the output files when it finishes?
      if (!gqSaveOutput)
        fprintf(scriptStream,"rm -f %s.o${JOB_ID}\n\n",gqScriptFilePath);
      
      fclose(scriptStream);
    }
  }
}

//

int
main(
  int       argc,
  char*     argv[]
)
{
  GQPEListRef   peList = NULL;
  GQPERef       thePE = NULL;
  int           argn,ch,optIndex;
  char*         scriptPath;
  int           status = 0;
  
  //  We need to do a first-time scan through the parameters to get
  //  the value for debug and to figure out if an input file was provided
  //  on the command line:
  argn = 1;
  while ( argn < argc ) {
    if ((strcmp(argv[argn],"--debug") == 0) || (strcmp(argv[argn],"-d") == 0)) {
      gqDebug = 1;
    }
    //  If there's no '-' heading-off the argument then we've found an
    //  input filename:
    else if (argv[argn][0] != '-') {
      //  Is it really a file?
      char*     canonicalPath = pathstr_canonicalPath(argv[argn]);
      
      if (canonicalPath) {
        if (pathstr_fileExists(canonicalPath)) {
          if (gqInputFilePath) free(gqInputFilePath);
          gqInputFilePath = canonicalPath;
          break;
        }
        free(canonicalPath);
      }
    }
    argn++;
  }
  
  //  First things first, get the p.e. configuration:
  if ((peList = GQPEListCreateWithXML(gqPEConfigPath)) == NULL) {
    printf("ERROR:  No gqueue p.e. configuration file at %s\n",gqPEConfigPath);
    GQExit(1);
  }
  
  //  Import any config params the user's provided in his/her environment:
  GQImportEnv();
  
  //  Scan all the c/l options:
  while ((ch = getopt_long(argc,argv,"vVqdQ:e:n:m:t:p:s:",gqOptions,&optIndex)) != -1) {
    switch (ch) {
    
      case 'v':
        gqVerbose = 1;
        break;
        
      case 'V': {
        printf("%s ",argv[0]);
        printvers(stdout,PROGRAM_VERSION);
        printf("\n%s %s\n",__DATE__,__TIME__);
        return 0;
      }
      
      case 'q':
        gqVerbose = 0;
        break;
      
      case 'd':
        gqDebug = 1;
        //  Once we know we're gonna be in debug mode, why not show our
        //  p.e. list:
        printf(":: gqueue: parallel environment sets:\n");
        GQPEListSummarizeAll(peList);
        break;
      
      //
      // Queue script/output directory:
      //
      case 'Q':
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg)
            gqQueueDirectory = pathstr_canonicalPath(optarg);
        }
        break;
      
      //
      // Email address:
      //
      case 'e':
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg)
            gqEmailAddress = strdup(optarg);
        }
        break;
      
      //
      // Processor count:
      //
      case 'n': {
        unsigned int    nproc;
        
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg && (sscanf(optarg,"%u",&nproc) == 1))
            gqNProc = nproc;
        }
        break;
      }
      
      //
      // Memory size:
      //
      case 'm': {
        unsigned int    mem;
        
        if (optarg && strlen(optarg)) {
          int           nchar;
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if ( mem = GQParseMemorySpec(optarg) )
            gqMem = ( mem < 32 ? 32 : mem );
        }
        break;
      }
      
      //
      // PE type:
      //
      case 't':
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg) {
            strtolower(optarg);
            if ((gqTargetPEList = GQPEListGetListWithIdentifier(peList,optarg)) == NULL) {
              printf("ERROR:  The PE type '%s' does not exist.\n",optarg);
              GQExit(1);
            }
          }
        }
        break;
      
      //
      // Priority:
      //
      case 'p':
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg) {
            if ( strcasecmp(optarg,gqPriority_Low) == 0 ) {
              gqPriority = gqPriority_Low;
            } else if ( strcasecmp(optarg,gqPriority_Medium) == 0 ) {
              gqPriority = gqPriority_Medium;
            } else if ( strcasecmp(optarg,gqPriority_High) == 0 ) {
              gqPriority = gqPriority_High;
            }
          }
        }
        break;
      
      //
      // What shell should the script run in:
      //
      case 's':
        if (optarg && strlen(optarg)) {
          while (*optarg && ( *optarg == '=' || isspace(*optarg) )) optarg++;
          if (*optarg)
            gqShellPath = strdup(optarg);
        }
        break;
    
    }
  }
  
  //  Did the user request help?  If so, display usage and just
  //  exit:
  if (gqHelp) {
    usage(argv[0]);
    return 0;
  }
  
  //  Did the user request a PE summary?
  if (gqPESummary) {
    //  Was a specific PE type selected?
    if (gqTargetPEList) {
      printf("Selected parallel-environment type:\n\n");
      GQPEListSummarize(gqTargetPEList);
    } else {
      printf("Available parallel-environment types (first is default):\n\n");
      GQPEListSummarizeAll(peList);
    }
    GQExit(0);
  }
  
  //  If no target pe-type was given, select the default now:
  if (!gqTargetPEList)
    gqTargetPEList = peList;
  
  //  If there was no input file, try to read from stdin:
  if (gqInputFilePath == NULL)
    gqInputFilePath = GQReadInputFileFromStdin();
  
  //  Get the queue directory setup:
  if (!gqQueueDirectory)
    gqQueueDirectory = strdup(gqDefaultQScriptDir);
  if ((gqQueueDirectory = GQSetupScriptDirectory()) == NULL)
    GQExit(1);
  
  //  Construct the queue script filepath:
  if ((gqScriptFilePath = GQCreateScriptFile()) == NULL)
    GQExit(1);
  
  //  Now's the time.  Do the pre-process script if it's there:
  if (scriptPath = GQPEListMakeScriptPath(gqTargetPEList,kGQPEListPreMergeScript)) {
    unsigned int      savedNProc = gqNProc;
    int               result;
    
    GQExportEnv();
    if (result = GQRunScript(scriptPath,gqTrue)) {
      printf("ERROR:  %s returned error code %d\n",scriptPath,result);
      GQExit(result);
    }
    free(scriptPath);
    
    //  If the user had already provided a processor count then that overrides
    //  what comes back from the environment import:
    if (savedNProc)
      gqNProc = savedNProc;
  }
  
  //  If the requested processor count is STILL zero then set it to 1:
  if (gqNProc == 0)
    gqNProc = 1;
    
  //  Get a parallel environment from the target PE list:
  if (!GQPEListUseImplicitSingle(gqTargetPEList) || (gqNProc > 1)) {
    if ((gqTargetPE = GQPEListSelectPE(gqTargetPEList,&gqNProc)) == NULL) {
      printf("ERROR: No parallel environment available for %u processor%s\n",gqNProc,(gqNProc != 1 ? "s:" : ":"));
      GQPEListSummarize(gqTargetPEList);
      GQExit(1);
    }
    if (gqVerbose)
      logtostream(stdout,"Selected parallel environment '%s' with nproc = %u",GQPEGetIdentifier(gqTargetPE),gqNProc);
      //  Do the PE validation script now:
    if (scriptPath = GQPEListMakeScriptPath(gqTargetPEList,kGQPEListValidatePEScript)) {
      int               result;
      
      GQExportEnv();
      if (result = GQRunScript(scriptPath,gqTrue)) {
        printf("ERROR:  %s returned error code %d\n",scriptPath,result);
        GQExit(result);
      }
      free(scriptPath);
    }
  } else if (gqVerbose)
    logtostream(stdout,"Single processor, no parallel environment selected");
  
  //  Add the SGE commands to the script:
  GQScriptSGESection();
  
  //  Time to merge the input file:
  GQScriptMergeInput();
  
  //  Do the post-process script if it's there:
  if (scriptPath = GQPEListMakeScriptPath(gqTargetPEList,kGQPEListPostMergeScript)) {
    int       result;
    
    GQExportEnv();
    if (result = GQRunScript(scriptPath,gqFalse)) {
      printf("ERROR:  %s returned error code %d\n",scriptPath,result);
      GQExit(result);
    }
    free(scriptPath);
  }
  
  //  Add the footer to the script:
  GQScriptFileFooter();
  
  //  Submit it!
  if (gqScriptFilePath) {
    char    commandBuffer[24];
    
    snprintf(commandBuffer,23,"qsub -p %d ",GQIntegerPriorityForString(gqPriority));
    if (scriptPath = strcatm(commandBuffer,gqScriptFilePath,NULL)) {
      if ( gqDebug )
        fprintf(stderr,":: system(\"%s\");\n",scriptPath);
      status = system(scriptPath);
      free(scriptPath);
    } else
      status = ENOMEM;
  } else
    status = ENOENT;
  GQExit(status);
}
