
#include "sll.h"
#include <stdio.h>
#include <string.h>

enum {
  sllNode_internalData    = 1 << 0,
  sllNode_freeWhenDone    = 1 << 1
};

typedef struct _sllNode {
  struct _sllNode*    link;
  short int           references;
  short int           flags;
  void*               data;
  size_t              dataLength;
  char*               id;
} sllNode;

//

sllNode*
__sllNode_alloc(
  size_t      extraBytes
)
{
  sllNode*    node = (sllNode*)malloc(sizeof(sllNode) + extraBytes);
  
  if (node) {
    node->link        = NULL;
    
    node->references  = 1;
    if (extraBytes) {
      node->flags     = sllNode_internalData;
      node->data      = ((void*)node) + sizeof(sllNode);
    } else {
      node->flags     = 0;
      node->data      = NULL;
    }
    node->dataLength  = extraBytes;
    
    node->id          = NULL;
  }
  return node;
}

//

void
__sllNode_dealloc(
  sllNode*      node
)
{
  if (node->link)
    sllNode_Release(node->link);
  if (node->data && !(node->flags & sllNode_internalData) && (node->flags & sllNode_freeWhenDone))
    free(node->data);
  if (node->id)
    free(node->id);
  free(node);
}

//
#pragma mark -
//

sllNodeRef
sllNode_Create()
{
  return __sllNode_alloc(0);
}

//

sllNodeRef
sllNode_CreateWithCapacity(
  size_t      bytes
)
{
  return __sllNode_alloc(bytes);
}

//

sllNodeRef
sllNode_CreateWithBuffer(
  void*       buffer,
  size_t      length
)
{
  sllNodeRef  node = __sllNode_alloc(length);
  
  if (node && length && buffer)
    memcpy(node->data,buffer,length);
  return node;
}

//

sllNodeRef
sllNode_CreateWithBufferNoCopy(
  void*       buffer,
  size_t      length,
  int         freeWhenDone
)
{
  sllNodeRef  node = NULL;
  
  if (length && buffer) {
    if (node = __sllNode_alloc(0)) {
      node->data          = buffer;
      node->dataLength    = length;
      if (freeWhenDone)
        node->flags |= sllNode_freeWhenDone;
    }
  }
  return node;
}

//

short int
sllNode_GetReferenceCount(
  sllNodeRef    aNode
)
{
  if (aNode)
    return aNode->references;
}

//

sllNodeRef
sllNode_Retain(
  sllNodeRef    aNode
)
{
  if (aNode)
    aNode->references++;
  return aNode;
}

void
sllNode_Release(
  sllNodeRef    aNode
)
{
  if (aNode) {
    if (--aNode->references == 0)
      __sllNode_dealloc(aNode);
  }
}

//

sllNodeRef
sllNode_GetLink(
  sllNodeRef    aNode
)
{
  return aNode->link;
}

//

void
sllNode_SetLink(
  sllNodeRef    aNode,
  sllNodeRef    newLink
)
{
  if (newLink)
    newLink = sllNode_Retain(newLink);
  if (aNode->link)
    sllNode_Release(aNode->link);
  aNode->link = newLink;
}

//

size_t
sllNode_GetDataLength(
  sllNodeRef    aNode
)
{
  return aNode->dataLength;
}

//

const void*
sllNode_GetDataPtr(
  sllNodeRef    aNode
)
{
  return aNode->data;
}

//

void
sllNode_SetDataPtr(
  sllNodeRef    aNode,
  const void*   aPtr
)
{
  if (!(aNode->flags & sllNode_internalData))
    aNode->data = (void*)aPtr;
}

//

void*
sllNode_GetMutableDataPtr(
  sllNodeRef    aNode
)
{
  if (aNode->flags & sllNode_internalData)
    return aNode->data;
  return NULL;
}

//

const char*
sllNode_GetIdentifier(
  sllNodeRef    aNode
)
{
  return aNode->id;
}

//

void
sllNode_SetIdentifier(
  sllNodeRef    aNode,
  const char*   newID
)
{
  if (aNode->id)
    free(aNode->id);
  if (newID)
    aNode->id = strdup(newID);
  else
    aNode->id = NULL;
}

//
#pragma mark -
//

unsigned int
sllNode_IndexOfNode(
  sllNodeRef    headNode,
  sllNodeRef    altNode
)
{
  unsigned int  index = 0;
  
  while ( headNode && headNode != altNode )
    headNode = headNode->link;
  if ( headNode )
    return index;
  return 0;
}

//

sllNodeRef
sllNode_PrependNode(
  sllNodeRef    headNode,
  sllNodeRef    altNode
)
{
  sllNode_SetLink(altNode,headNode);
  //  Account for a change in head node:
  if (headNode)
    sllNode_Release(headNode);
  return sllNode_Retain(altNode);
}

//

sllNodeRef
sllNode_AppendNode(
  sllNodeRef    headNode,
  sllNodeRef    altNode
)
{
  if (headNode) {
    sllNodeRef  saved = headNode;
    
    while (headNode->link)
      headNode = headNode->link;
    sllNode_SetLink(headNode,altNode);
    return saved;
  }
  return sllNode_Retain(altNode);
}

//

sllNodeRef
sllNode_InsertNodeAtIndex(
  sllNodeRef    headNode,
  sllNodeRef    altNode,
  unsigned int  index
)
{
  sllNodeRef    saved = headNode;
  
  if (index == 0) {
    //  Prepend operation:
    return sllNode_PrependNode(headNode,altNode);
  }
  index--;
  while (headNode && index--)
    headNode = headNode->link;
  //  headNode -> node AFTER WHICH the new node should be found:
  if (headNode) {
    sllNode_SetLink(altNode,headNode->link);
    sllNode_SetLink(headNode,altNode);
  }
  return saved;
}

//

sllNodeRef
sllNode_SortedInsertion(
  sllNodeRef              headNode,
  sllNodeRef              altNode,
  sllNodeCompareCallback  compare,
  const void*             context,
  int                     allowDuplicates
)
{
  sllNodeRef    saved = headNode;
  
  if (headNode) {
    sllNodeRef  prevNode = NULL;
    
    while (headNode) {
      //  As long as we're greater than the headNode, keep
      //  going:
      int       cmpResult = compare(altNode,headNode,context);
      
      if (cmpResult < 0) {
        //  altNode belongs before the current node:
        sllNode_SetLink(altNode,headNode);
        if (prevNode) {
          sllNode_SetLink(prevNode,altNode);
          return saved;
        } else {
          //  altNode is the new list head; account for the change
          //  if necessary:
          if (saved)
            sllNode_Release(saved);
          return sllNode_Retain(altNode);
        }
      } else if (cmpResult == 0 && !allowDuplicates) {
        //  Exactly equal to the current node and we don't allow duplicates,
        //  so replace it!
        sllNode_SetLink(altNode,headNode->link);
        if (prevNode) {
          sllNode_SetLink(prevNode,altNode);
          return saved;
        } else {
          sllNode_Release(saved);
          return sllNode_Retain(altNode);
        }
      }
      prevNode = headNode;
      headNode = headNode->link;
    }
    //  We've gotten all the way to the end of the list; append
    //  the record:
    sllNode_SetLink(prevNode,altNode);
    return saved;
  }
  return sllNode_Retain(altNode);
}

//

sllNodeRef
sllNode_RemoveNodeAtIndex(
  sllNodeRef    headNode,
  unsigned int  index
)
{
  sllNodeRef    saved = headNode;
  
  if (headNode) {
    sllNodeRef  prevNode = NULL;
    
    if (index == 0) {
      //  head node is being removed!
      prevNode = sllNode_Retain(headNode->link);
      sllNode_Release(headNode);
      return prevNode;
    }
    while (headNode && index--) {
      prevNode = headNode;
      headNode = headNode->link;
    }
    if (headNode) {
      sllNodeRef    newLink = sllNode_Retain(headNode->link);
      
      sllNode_SetLink(prevNode,newLink);
      sllNode_Release(newLink);
    }
    return saved;
  }
  return NULL;
}

//

sllNodeRef
sllNode_RemoveNode(
  sllNodeRef    headNode,
  sllNodeRef    altNode
)
{
  sllNodeRef    saved = headNode,prevNode;
  
  if (headNode == altNode) {
    //  head node is being removed!
    saved = sllNode_Retain(headNode->link);
    sllNode_Release(headNode);
    return saved;
  }
  do {
    prevNode = headNode;
    if ((headNode = headNode->link) == altNode) {
      sllNodeRef    newLink = sllNode_Retain(headNode->link);
      
      sllNode_SetLink(prevNode,newLink);
      sllNode_Release(newLink);
      break;
    }
  } while (headNode);
  return saved;
}

//

unsigned int
sllNode_GetCount(
  sllNodeRef    headNode
)
{
  unsigned int  count = 0;
  
  while (headNode) {
    count++;
    headNode = headNode->link;
  }
  return count;
}

//

sllNodeRef
sllNode_GetNodeAtIndex(
  sllNodeRef    headNode,
  unsigned int  index
)
{
  while (headNode && index--)
    headNode = headNode->link;
  return headNode;
}

//

sllNodeRef
sllNode_GetNodeByIdentifier(
  sllNodeRef    headNode,
  const char*   id,
  int           caseSensitive
)
{
  if (caseSensitive) {
    while (headNode) {
      if (headNode->id && strcmp(headNode->id,id) == 0)
        return headNode;
      headNode = headNode->link;
    }
  } else {
    while (headNode) {
      if (headNode->id && strcasecmp(headNode->id,id) == 0)
        return headNode;
      headNode = headNode->link;
    }
  }
  return NULL;
}

//

sllNodeRef
sllNode_Search(
  sllNodeRef            headNode,
  sllNodeSearchCallback search,
  const void*           context
)
{
  while (headNode) {
    if (search(headNode,context))
      return headNode;
    headNode = headNode->link;
  }
  return NULL;
}

//

sllNodeRef
sllNode_Summarize(
  sllNodeRef  headNode
)
{
  while (headNode) {
    if (headNode->id)
      printf("%s[%u] -> ",headNode->id,headNode->references);
    else
      printf("%p[%u] -> ",headNode,headNode->references);
    headNode = headNode->link;
  }
  printf("EOL\n");
}

//
#pragma mark -
//

int
sllNodeIdentifierSearcher(
  sllNodeRef    node,
  const void*   context
)
{
  return ( strcmp(node->id,(char*)context) == 0 );
}

int
sllNodeIdentifierComparator(
  sllNodeRef    node1,
  sllNodeRef    node2,
  const void*   context
)
{
  return strcmp(node1->id,node2->id);
}
