/*!
  @header sll.h
  Implements a generalized singly-linked list (hence, sll) data type with reference counting
  for the nodes.
  @copyright Jeffrey T. Frey / NSS / University of Delaware
  @updated 2006-01-13
*/

#ifndef SLL_H
#define SLL_H

#include <stdlib.h>

/*!
  @typedef sllNodeRef
  Opaque reference to a sll node.
*/
typedef struct _sllNode* sllNodeRef;

/*!
  @typedef sllNodeSearchCallback
  Type for a callback function to be used when searching a sll for a particular
  node.
  @param node Node to examine
  @param context Passed to sllNode_Search() and subsequently to the callback
    function each time it is called; use this pointer-sized argument to pass search
    criteria to the callback function.
  @result A non-zero value should be returned if node is a match.
*/
typedef int (*sllNodeSearchCallback)(sllNodeRef node,const void* context);

/*!
  @typedef sllNodeCompareCallback
  Type for a callback function to be used when comparing two sll nodes during
  sorted-insertion.
  @param node1 A sll node
  @param node2 Another sll node
  @param context Passed to sllNode_SortedInsertion() and subsequently to the callback
    function each time it is called.
  @result An integer greater than, equal to, or less than 0, according to whether
    node1 is greater than, equal to, or less than the node2.
*/
typedef int (*sllNodeCompareCallback)(sllNodeRef node1,sllNodeRef node2,const void* context);

int sllNodeIdentifierSearcher(sllNodeRef node,const void* context);
int sllNodeIdentifierComparator(sllNodeRef node1,sllNodeRef node2,const void* context);

sllNodeRef sllNode_Create();
sllNodeRef sllNode_CreateWithCapacity(size_t bytes);
sllNodeRef sllNode_CreateWithBuffer(void* buffer,size_t length);
sllNodeRef sllNode_CreateWithBufferNoCopy(void* buffer,size_t length,int freeWhenDone);

short int sllNode_GetReferenceCount(sllNodeRef aNode);
sllNodeRef sllNode_Retain(sllNodeRef aNode);
void sllNode_Release(sllNodeRef aNode);

sllNodeRef sllNode_GetLink(sllNodeRef aNode);
void sllNode_SetLink(sllNodeRef aNode,sllNodeRef newLink);

size_t sllNode_GetDataLength(sllNodeRef aNode);
const void* sllNode_GetDataPtr(sllNodeRef aNode);
void sllNode_SetDataPtr(sllNodeRef aNode,const void* aPtr);
void* sllNode_GetMutableDataPtr(sllNodeRef aNode);

const char* sllNode_GetIdentifier(sllNodeRef aNode);
void sllNode_SetIdentifier(sllNodeRef aNode,const char* newID);

sllNodeRef sllNode_PrependNode(sllNodeRef headNode,sllNodeRef altNode);
sllNodeRef sllNode_AppendNode(sllNodeRef headNode,sllNodeRef altNode);
sllNodeRef sllNode_InsertNodeAtIndex(sllNodeRef headNode,sllNodeRef altNode,unsigned int index);
sllNodeRef sllNode_SortedInsertion(sllNodeRef headNode,sllNodeRef altNode,
              sllNodeCompareCallback compare,const void* context,int allowDuplicates);
sllNodeRef sllNode_RemoveNodeAtIndex(sllNodeRef headNode,unsigned int index);
sllNodeRef sllNode_RemoveNode(sllNodeRef headNode,sllNodeRef altNode);

unsigned int sllNode_GetCount(sllNodeRef headNode);
unsigned int sllNode_IndexOfNode(sllNodeRef headNode,sllNodeRef altNode);
sllNodeRef sllNode_GetNodeAtIndex(sllNodeRef headNode,unsigned int index);
sllNodeRef sllNode_GetNodeByIdentifier(sllNodeRef headNode,const char* id,int caseSensitive);
sllNodeRef sllNode_Search(sllNodeRef headNode,sllNodeSearchCallback search,const void* context);

#endif
